"""Exception classes used in TL2cgen"""


class TL2cgenError(Exception):
    """Error thrown by TL2cgen"""


class TL2cgenLibraryNotFound(Exception):
    """Error thrown by when TL2cgen is not found"""
