/*!
 * Copyright (c) 2023 by Contributors
 * \file header_template.h
 * \author Hyunsu Cho
 * \brief Template for header
 */

#ifndef TL2CGEN_DETAIL_COMPILER_TEMPLATES_HEADER_TEMPLATE_H_
#define TL2CGEN_DETAIL_COMPILER_TEMPLATES_HEADER_TEMPLATE_H_

namespace tl2cgen::compiler::detail::templates {

extern char const* const query_functions_prototype_template;
extern char const* const header_template;

}  // namespace tl2cgen::compiler::detail::templates

#endif  // TL2CGEN_DETAIL_COMPILER_TEMPLATES_HEADER_TEMPLATE_H_
