/*!
 * Copyright (c) 2023 by Contributors
 * \file code_folder_template.h
 * \author Hyunsu Cho
 * \brief Template for evaluation logic for folded code
 */

#ifndef TL2CGEN_DETAIL_COMPILER_TEMPLATES_CODE_FOLDER_TEMPLATE_H_
#define TL2CGEN_DETAIL_COMPILER_TEMPLATES_CODE_FOLDER_TEMPLATE_H_

namespace tl2cgen::compiler::detail::templates {

extern char const* const eval_loop_template;
extern char const* const eval_loop_template_without_categorical_feature;

}  // namespace tl2cgen::compiler::detail::templates

#endif  // TL2CGEN_DETAIL_COMPILER_TEMPLATES_CODE_FOLDER_TEMPLATE_H_
